import codecs
import errno
import shutil
import os
import json
from tqdm import tqdm
import bcolz
import pandas
import numpy as np

from brine.exceptions import BrineError
from brine.schema import Schema, Integer, Float, Category, String, Image, IntegerArray, FloatArray, CategoryArray


class Builder(object):

    def build_from_config(self, config_file_path, destination_dir_path):

        try:
            with codecs.open(config_file_path, 'r', encoding='utf-8') as f:
                config = json.loads(f.read())
                config_columns = config['columns']
                config_path = config['path']
                config_extra_data = config.get('extra_data') or None
        except (IOError, ValueError, KeyError):
            raise BuilderError

        dtype = {}
        converters = {}
        num_image_columns = 0
        for config_column in config_columns:
            name = config_column['name']
            column_type = config_column['type']
            if column_type == 'integer':
                dtype[name] = np.int64
            elif column_type == 'float':
                dtype[name] = np.float
            elif column_type == 'category':
                converter = CategoryConverter()
                converters[name] = converter
            elif column_type == 'string':
                dtype[name] = np.unicode
            elif column_type == 'image':
                dtype[name] = np.unicode
                num_image_columns = num_image_columns + 1
            elif column_type == 'integer_array':
                converter = ArrayConverter(np.int64)
                converters[name] = converter
            elif column_type == 'float_array':
                converter = ArrayConverter(np.float)
                converters[name] = converter
            elif column_type == 'category_array':
                converter = ArrayConverter(CategoryConverter())
                converters[name] = converter

        csv_file_path = os.path.join(os.path.dirname(config_file_path), config_path)
        try:
            df = pandas.read_csv(csv_file_path, dtype=dtype, converters=converters)
        except IOError:
            raise BuilderError
        df = df.reindex(columns=[config_column['name'] for config_column in config_columns])
        progress_bar = tqdm(unit=' images', total=num_image_columns * df.shape[0])

        schema = Schema()
        for config_column in config_columns:
            name = config_column['name']
            column_type = config_column['type']
            if column_type == 'integer':
                schema.add_column(name, Integer())
            elif column_type == 'float':
                schema.add_column(name, Float())
            elif column_type == 'category':
                schema.add_column(name, Category(converters[name].categories))
            elif column_type == 'string':
                schema.add_column(name, String())
            elif column_type == 'image':
                schema.add_column(name, Image())
                for file_path in df[name]:
                    self.copy_image_file(file_path, os.path.dirname(config_file_path), destination_dir_path)
                    progress_bar.update(1)
            elif column_type == 'integer_array':
                schema.add_column(name, IntegerArray())
            elif column_type == 'float_array':
                schema.add_column(name, FloatArray())
            elif column_type == 'category_array':
                schema.add_column(name, CategoryArray(converters[name].items_converter.categories))

        ctable = bcolz.ctable.fromdataframe(df, rootdir=os.path.join(destination_dir_path, 'bcolz'))
        ctable.attrs['extra_data'] = json.dumps(config_extra_data)
        ctable.attrs['schema'] = json.dumps(schema.to_obj())
        ctable.flush()

    def copy_image_file(self, file_path, src_dir_path, dst_dir_path):
        src_file_path = os.path.join(src_dir_path, file_path)
        dst_file_path = os.path.join(dst_dir_path, 'images', file_path)
        if not os.path.abspath(dst_file_path).startswith(dst_dir_path):
            raise BuilderError
        try:
            shutil.copyfile(src_file_path, dst_file_path)
        except IOError as ex:
            if ex.errno != errno.ENOENT:
                raise BuilderError
            try:
                os.makedirs(os.path.dirname(dst_file_path))
                shutil.copyfile(src_file_path, dst_file_path)
            except IOError:
                raise BuilderError


class BuilderError(BrineError):
    pass


class CategoryConverter(object):

    def __init__(self):
        self.categories_dict = {}

    def __call__(self, value):
        if value not in self.categories_dict:
            self.categories_dict[value] = len(self.categories_dict)
        return self.categories_dict[value]

    @property
    def categories(self):
        return list(self.categories_dict)


class ArrayConverter(object):

    def __init__(self, items_converter):
        self.items_converter = items_converter

    def __call__(self, value):
        return [self.items_converter(x) for x in value.split(' ')]



import argparse
import sys

from brine.install import install
from brine.uninstall import uninstall
from brine.ls import ls
from brine.info import info
from brine.build import build
from brine.push import push
from brine.exceptions import BrineError


def main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(title='command', dest='command')
    subparsers.required = True

    # brine install <dataset>
    install_parser = subparsers.add_parser('install')
    install_parser.add_argument(
        'dataset',
        metavar='<dataset>',
        type=str)
    install_parser.set_defaults(func=lambda args: install(args.dataset))

    # brine uninstall <dataset>
    uninstall_parser = subparsers.add_parser('uninstall')
    uninstall_parser.add_argument(
        'dataset',
        metavar='<dataset>',
        type=str)
    uninstall_parser.set_defaults(func=lambda args: uninstall(args.dataset))

    # brine list
    ls_parser = subparsers.add_parser('list')
    ls_parser.set_defaults(func=lambda args: ls())

    # brine info <dataset>
    info_parser = subparsers.add_parser('info')
    info_parser.add_argument(
        'dataset',
        metavar='<dataset>',
        type=str)
    info_parser.set_defaults(func=lambda args: info(args.dataset))

    # brine build <dataset> --config=<config>
    build_parser = subparsers.add_parser('build')
    build_parser.add_argument(
        'dataset',
        metavar='<dataset>',
        type=str)
    build_parser.add_argument(
        '--config',
        metavar='<config>',
        required=True,
        type=str)
    build_parser.set_defaults(func=lambda args: build(args.dataset, args.config))

    # brine push <dataset>
    # push_parser = subparsers.add_parser('push')
    # push_parser.add_argument(
    #     'dataset',
    #     metavar='<dataset>',
    #     type=str)
    # push_parser.set_defaults(func=lambda args: push(args.dataset))

    args = parser.parse_args()

    try:
        args.func(args)
    except BrineError as ex:
        print('Error: %s' % str(ex), file=sys.stderr)
        sys.exit(1)

    sys.exit(0)



import os
from brine.api import get_version_for_install
from brine.dataset_manager import DatasetManager
from brine.download import download_and_extract
from brine.util import TemporaryDirectory


def install(dataset_name):
    dataset_manager = DatasetManager.get_from_dir(dataset_name, os.getcwd())
    dataset_manager.check_can_install()

    response = get_version_for_install(dataset_name)

    with TemporaryDirectory() as temp_dir_path:
        download_and_extract(response['signed_url'], temp_dir_path)
        dataset_manager.create_from_dir(temp_dir_path, response['version_number'])

    print('Dataset %s (v%s) was installed.' % (dataset_name, dataset_manager.version()))



import os
from brine.dataset_manager import DatasetManager


def ls():
    for dataset_manager in DatasetManager.list_in_dir(os.getcwd()):
        version = dataset_manager.version()
        if version is None:
            print('%s (local)' % dataset_manager.name)
        else:
            print('%s (v%s)' % (dataset_manager.name, version))



import os
from brine.api import create_upload_session, get_upload_session_signed_url, complete_upload_session, ApiError
from brine.dataset_manager import DatasetManager, DatasetManagerError
from brine.util import TemporaryDirectory
from brine.upload import upload, UploadError
from brine.archive import archive, ArchiveError


def push(dataset_name):
    dataset_manager = DatasetManager.get_from_dir(dataset_name, os.getcwd())
    dataset_manager.check_can_push()

    response = create_upload_session(dataset_name)
    upload_session_id = response['upload_session_id']

    with TemporaryDirectory() as temp_dir_path:

        tar_file_path = os.path.join(temp_dir_path, 'dataset.tar')

        archive(dataset_manager.path, tar_file_path)

        response = get_upload_session_signed_url(upload_session_id, 'dataset.tar')
        signed_url = response['signed_url']

        upload(signed_url, tar_file_path)

    response = complete_upload_session(upload_session_id)
    dataset_manager.set_version(response['version_number'])

    print('Dataset %s (v%s) was pushed.' % (dataset_manager.name, dataset_manager.version()))



import codecs
import hashlib
import os
import re
import requests
from tqdm import tqdm

from brine.exceptions import BrineError


RETRY_STATUS_CODES = [408, 429, 500, 502, 503, 504]
BUFFER_SIZE = 8192


def upload(signed_url, source_file_path):
    session_url = _get_session_url(signed_url)

    try:
        with codecs.open(source_file_path, 'rb') as f:
            f.seek(0, os.SEEK_END)
            file_length = f.tell()
            upload_progress = UploadProgress(total=file_length)
            while not _upload_attempt(session_url, f, file_length, upload_progress):
                pass
    except IOError:
        raise UploadError('Failed to upload file to server.')


def _get_session_url(signed_url):
    r = requests.post(signed_url, headers={
        'Content-Length': '0',
        'Content-Type': 'application/octet-stream',
        'x-goog-resumable': 'start',
    })
    if r.status_code != 201:
        raise UploadError('Failed to create a signed upload URL.')
    return r.headers['Location']


def _upload_attempt(session_url, f, file_length, upload_progress):
    try:
        r = requests.put(session_url, headers={
            'Content-Length': '0',
            'Content-Range': 'bytes */%d' % file_length,
        })
    except requests.RequestException:
        return False

    if r.status_code in RETRY_STATUS_CODES:
        return False
    elif r.status_code in (200, 201):
        # handle edge case for file upload already completed
        etag = r.headers['ETag'].strip('"')
        if etag != upload_progress.md5.hexdigest():
            raise UploadError('Failed to upload file to server.')
        return True
    elif r.status_code != 308:
        raise UploadError('Failed to upload file to server.')

    if 'Range' not in r.headers:
        uploaded_range = (0, 0)
        headers = {}
    else:
        # this byte range header is inclusive, so bytes=(0, 0) means that one byte is uploaded
        m = re.search(r'bytes=(\d+)-(\d+)', r.headers['Range'])
        # convert to exclusive range
        uploaded_range = (int(m.group(1)), int(m.group(2)) + 1)
        headers = {
            'Content-Range': 'bytes %d-%d/%d' % (
                uploaded_range[1],
                file_length - 1,
                file_length,
            )
        }
    try:
        r = requests.put(
            session_url,
            headers=headers,
            data=upload_progress.read(f, uploaded_range[1])
        )
    except requests.RequestException:
        return False

    if r.status_code in RETRY_STATUS_CODES:
        return False
    elif r.status_code in (200, 201):
        etag = r.headers['ETag'].strip('"')
        if etag != upload_progress.md5.hexdigest():
            raise UploadError('Failed to upload file to server.')
        return True
    else:
        raise UploadError('Failed to upload file to server.')


class UploadProgress(object):

    def __init__(self, total):
        self.offset = 0
        self.md5 = hashlib.md5()
        self.progress_bar = tqdm(unit='B', unit_scale=True, total=total)

    def read(self, f, offset):
        f.seek(offset)
        while True:
            chunk = f.read(BUFFER_SIZE)
            if not chunk:
                return
            self.update(chunk, offset)
            offset += len(chunk)
            yield chunk

    def update(self, chunk, offset):
        if self.offset >= len(chunk) + offset:
            return
        new_chunk = chunk[self.offset - offset:]
        self.md5.update(new_chunk)
        self.progress_bar.update(len(new_chunk))
        self.offset = len(chunk) + offset


class UploadError(BrineError):
    pass



import numpy as np
import threading
from keras.preprocessing.image import img_to_array
from brine.iterator import Iterator


class KerasGenerator(object):
    """A generator that yields batches of samples. Can be used with Keras' `fit_generator` and `predict_generator`.

    Refer to documentation for :meth:`brine.dataset.Dataset.to_keras`.
    """
    def __init__(self, dataset, x_column, y_column=None, batch_size=32,
                 shuffle=True, seed=None, processing_function=None):
        self.dataset = dataset
        self.x_column = getattr(dataset.columns, x_column)
        if y_column is not None:
            self.y_column = getattr(dataset.columns, y_column)
        else:
            self.y_column = None
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.processing_function = processing_function
        self.lock = threading.Lock()
        self.index_generator = Iterator().flow_index(len(self.dataset),
                                                     batch_size=batch_size, shuffle=shuffle, seed=seed)

    def __iter__(self):
        return self

    def next(self):
        with self.lock:
            index_array, current_index, current_batch_size = next(self.index_generator)
        xs = []
        ys = []
        for i, index in enumerate(index_array):
            row = self.dataset[index]
            x = getattr(row, self.x_column.name)
            x = self._preprocess_field(self.x_column, x)

            if self.y_column is not None:
                y = getattr(row, self.y_column.name)
                y = self._preprocess_field(self.y_column, y)
            else:
                y = None

            if self.processing_function is not None:
                data = self.processing_function((x, y))

            x = data[0]
            xs.append(x)
            if self.y_column is not None:
                y = data[1]
                ys.append(y)

        if self.y_column is not None:
            return np.stack(xs), np.stack(ys)
        else:
            return np.stack(xs)

    def steps_per_epoch(self):
        """The number of batches in one full epoch.

        Equal to the number of rows in the dataset divided by the batch size.

        Returns
        -------
        int
        """
        dataset_len = len(self.dataset)
        if dataset_len % self.batch_size == 0:
            return int(dataset_len/self.batch_size)
        else:
            return (int(dataset_len/self.batch_size) + 1)

    def __next__(self, *args, **kwargs):
        return self.next(*args, **kwargs)

    def _preprocess_field(self, column, value):
        if column.isimage():
            value = img_to_array(self.dataset.load_image(value))
        elif column.categories is not None:
            value = self._get_category_index(column, value)
        return value

    def _get_category_index(self, column, value):
        # TODO(rohan): Use a dict for this
        if isinstance(value, tuple):
            return tuple(map(lambda v: column.categories.index(v), value))
        else:
            return column.categories.index(value)



import reprlib
from brine.exceptions import BrineError


class Schema(object):

    def __init__(self):
        self.columns = []

    def __repr__(self):
        return 'Schema(columns=%s)' % repr(self.columns)

    def add_column(self, name, column_type):
        self.append_column(Column(name, column_type))

    def append_column(self, column):
        column.set_index(len(self.columns))
        self.columns.append(column)

    def to_obj(self):
        return [column.to_obj() for column in self.columns]

    @classmethod
    def from_obj(cls, obj):
        schema = cls()
        for column_obj in obj:
            schema.append_column(Column.from_obj(column_obj))
        return schema


class Column(object):

    def __init__(self, name, column_type):
        self.name = name
        self.column_type = column_type
        self.index = None

    def __repr__(self):
        if self.categories:
            return 'Column(name=%s, type=%s, categories=%s)' % (
                self.name, reprlib.repr(self.column_type), reprlib.repr(self.categories))
        else:
            return 'Column(name=%s, type=%s)' % (self.name, repr(self.column_type))

    def isarray(self):
        return isinstance(self.column_type, (IntegerArray, FloatArray, CategoryArray))

    @property
    def categories(self):
        if not isinstance(self.column_type, (Category, CategoryArray)):
            return None
        return self.column_type.categories

    def isimage(self):
        return isinstance(self.column_type, Image)

    def set_index(self, index):
        self.index = index

    def to_obj(self):
        if isinstance(self.column_type, Integer):
            return {'name': self.name, 'type': 'integer'}
        elif isinstance(self.column_type, Float):
            return {'name': self.name, 'type': 'float'}
        elif isinstance(self.column_type, Category):
            return {'name': self.name, 'type': 'category', 'categories': self.column_type.categories}
        elif isinstance(self.column_type, String):
            return {'name': self.name, 'type': 'string'}
        elif isinstance(self.column_type, Image):
            return {'name': self.name, 'type': 'image'}
        elif isinstance(self.column_type, IntegerArray):
            return {'name': self.name, 'type': 'integer_array'}
        elif isinstance(self.column_type, FloatArray):
            return {'name': self.name, 'type': 'float_array'}
        elif isinstance(self.column_type, CategoryArray):
            return {'name': self.name, 'type': 'category_array', 'categories': self.column_type.categories}

    @classmethod
    def from_obj(cls, obj):
        name = obj.get('name')
        column_type = obj.get('type')
        if column_type == 'integer':
            return cls(name, Integer())
        elif column_type == 'float':
            return cls(name, Float())
        elif column_type == 'category':
            categories = obj.get('categories')
            return cls(name, Category(categories))
        elif column_type == 'string':
            return cls(name, String())
        elif column_type == 'image':
            return cls(name, Image())
        elif column_type == 'integer_array':
            return cls(name, IntegerArray())
        elif column_type == 'float_array':
            return cls(name, FloatArray())
        elif column_type == 'category_array':
            categories = obj.get('categories')
            return cls(name, CategoryArray(categories))


class ColumnType(object):
    pass


class Integer(ColumnType):

    def __repr__(self):
        return 'Integer'


class Float(ColumnType):

    def __repr__(self):
        return 'Float'


class Category(ColumnType):

    def __init__(self, categories):
        self.categories = categories

    def __repr__(self):
        return 'Category'


class String(ColumnType):

    def __repr__(self):
        return 'String'


class Image(ColumnType):

    def __repr__(self):
        return 'Image'


class IntegerArray(ColumnType):

    def __repr__(self):
        return 'IntegerArray'


class FloatArray(ColumnType):

    def __repr__(self):
        return 'FloatArray'


class CategoryArray(ColumnType):

    def __init__(self, categories):
        self.categories = categories

    def __repr__(self):
        return 'CategoryArray'


class SchemaError(BrineError):
    pass



import tarfile
import os

from brine.env import Env
from brine.exceptions import BrineError


def archive(source_dir_path, tar_file_path):
    try:
        with tarfile.open(tar_file_path, 'w') as tar:
            tar.add(
                source_dir_path,
                arcname=os.path.sep,
                filter=lambda tarinfo: None if tarinfo.name == Env.HIDDEN_FILE_NAME else tarinfo)
    except tarfile.TarError:
        raise ArchiveError('Failed to archive file %s.' % tar_file_path)


class ArchiveError(BrineError):
    pass



import os
from brine.dataset_manager import DatasetManager
from brine.builder import Builder
from brine.util import TemporaryDirectory


def build(dataset_name, config_file_path):
    dataset_manager = DatasetManager.get_from_dir(dataset_name, os.getcwd())
    dataset_manager.check_can_install()

    builder = Builder()
    with TemporaryDirectory() as temp_dir_path:
        builder.build_from_config(config_file_path, temp_dir_path)
        dataset_manager.create_from_dir(temp_dir_path)

    print('Dataset %s was built.' % dataset_name)



import tarfile
import requests
from tqdm import tqdm

from brine.exceptions import BrineError


def download_and_extract(signed_url, destination_dir_path):
    try:
        _download_and_extract(signed_url, destination_dir_path)
    except requests.exceptions.RequestException:
        raise DownloadError('Unable to reach server. Please try again later.')
    except tarfile.TarError:
        raise DownloadError('Failed to extract file %s.' % destination_dir_path)


def _download_and_extract(signed_url, destination_dir_path):
    r = requests.get(signed_url, stream=True)
    r.raise_for_status()

    try:
        content_length = int(r.headers.get('Content-Length'))
    except (ValueError, TypeError):
        content_length = 0
    download_progress = DownloadProgress(r.raw, total=content_length)
    with tarfile.open(fileobj=download_progress, mode='r|') as tar:
        tar.extractall(path=destination_dir_path)


class DownloadError(BrineError):
    pass


class DownloadProgress(object):

    def __init__(self, fileobj, total):
        self.fileobj = fileobj
        self.progress_bar = tqdm(unit='B', unit_scale=True, total=total)

    def read(self, size=-1):
        data = self.fileobj.read(size)
        self.progress_bar.update(len(data))
        return data



import requests

from brine.env import Env
from brine.exceptions import BrineError


def get_dataset_for_info(dataset_name):
    query = '''
    query ($datasetName: String!) {
        dataset(name: $datasetName) {
            name
            description
            versions(first: 100) {
                edges {
                    node {
                        number
                    }
                }
            }
        }
    }
    '''
    variables = {'datasetName': dataset_name}
    data = _graphql(query, variables)
    try:
        return {
            'name': data['dataset']['name'],
            'description': data['dataset']['description'],
            'versions': list(map(lambda x: x['node']['number'], data['dataset']['versions']['edges'])),
        }
    except KeyError:
        raise ApiError('Invalid server response.')
    except TypeError:
        raise ApiError('Dataset %s does not exist.' % dataset_name)


def get_version_for_install(dataset_name):
    query = '''
    query ($datasetName: String!) {
        dataset(name: $datasetName) {
            latestVersion {
                number
                signedUrl
            }
        }
    }
    '''
    variables = {'datasetName': dataset_name}
    data = _graphql(query, variables)
    try:
        return {
            'version_number': data['dataset']['latestVersion']['number'],
            'signed_url': data['dataset']['latestVersion']['signedUrl'],
        }
    except KeyError:
        raise ApiError('Invalid server response.')
    except TypeError:
        raise ApiError('Dataset %s does not exist.' % dataset_name)


def create_upload_session(dataset_name):
    query = '''
    mutation ($datasetName: String!) {
        createUploadSession(input: {datasetName: $datasetName, clientMutationId: "1"}) {
            uploadSession {
                id
            }
        }
    }'''
    variables = {'datasetName': dataset_name}
    data = _graphql(query, variables)
    try:
        return {
            'upload_session_id': data['createUploadSession']['uploadSession']['id']
        }
    except KeyError:
        raise ApiError('Invalid server response.')


def get_upload_session_signed_url(upload_session_id, filename):
    query = '''
    query ($id: ID!, $filename: String!) {
        uploadSession(id: $id) {
            signedUrl(filename: $filename)
        }
    }'''
    variables = {'id': upload_session_id, 'filename': filename}
    data = _graphql(query, variables)
    try:
        return {
            'signed_url': data['uploadSession']['signedUrl']
        }
    except KeyError:
        raise ApiError('Invalid server response.')


def complete_upload_session(upload_session_id):
    query = '''
    mutation ($id: ID!) {
        completeUploadSession(input: {id: $id, clientMutationId: "1"}) {
            version {
                id
                number
            }
        }
    }'''
    variables = {'id': upload_session_id}
    data = _graphql(query, variables)
    try:
        return {
            'version_id': data['completeUploadSession']['version']['id'],
            'version_number': data['completeUploadSession']['version']['number'],
        }
    except KeyError:
        raise ApiError('Invalid server response.')


def _graphql(query, variables):
    payload = {'query': query, 'variables': variables}
    headers = {'Content-Type': 'application/json',
               'Accept': 'application/json'}
    try:
        r = requests.post(Env.ENDPOINT, json=payload, headers=headers)
        r.raise_for_status()
        return r.json()['data']
    except requests.exceptions.RequestException as ex:
        raise ApiError('Unable to reach server.')
    except KeyError:
        raise ApiError('Invalid server response.')


class ApiError(BrineError):
    pass



from torch.utils.data import Dataset


class ImagePathToImage(object):
    """Transform that converts an Image Path into a PIL image.
    """
    def __init__(self, dataset, filepaths):
        self.dataset = dataset
        self.filepaths = filepaths

    def __call__(self, row):
        images = {filepath: self.dataset.load_image(getattr(row, filepath)) for filepath in self.filepaths}
        return row._replace(**images)


class PytorchDataset(Dataset):
    """A PyTorch Dataset that returns rows from the Brine Dataset. Can be used anywhere a PyTorch Dataset is used,
    including with PyTorch DataLoaders.

    See documentation for :meth:`~brine.dataset.Dataset.to_pytorch`
    """
    def __init__(self, dataset, transform=None, transform_columns=None):
        self.dataset = dataset
        self.image_columns = {column.name for column in self.dataset.columns if column.isimage()}
        self.category_columns = {column for column in self.dataset.columns if column.categories is not None}
        if (transform_columns is not None) and (transform is None):
            raise "transform must be set if transform_columns is set"
        if transform is not None:
            self.transform = transform
            self.transform_columns = transform_columns
        else:
            self.transform = ImagePathToImage(dataset, self.image_columns)
            self.transform_columns = None

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, index):
        row = self.dataset[index]
        category_indexed = {column.name: self._get_category_index(column, getattr(row, column.name))
                            for column in self.category_columns}
        row = row._replace(**category_indexed)
        if self.transform_columns == 'images':
            for column_name in self.image_columns:
                image = self.dataset.load_image(getattr(row, column_name))
                row = row._replace(**{column_name: self.transform(image)})
        else:
            row = self.transform(row)
        return row

    def _get_category_index(self, column, value):
        # TODO(rohan): Use a dict for this
        if isinstance(value, tuple):
            return tuple(map(lambda v: column.categories.index(v), value))
        else:
            return column.categories.index(value)



import errno
import shutil
import tempfile


class TemporaryDirectory(object):

    def __init__(self):
        self.name = tempfile.mkdtemp()

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.name)

    def __enter__(self):
        return self.name

    def __exit__(self, exc, value, tb):
        try:
            shutil.rmtree(self.name)
        except OSError as ex:
            if ex.errno != errno.ENOENT:
                raise



# Taken and modifed from https://github.com/fastai/courses/blob/master/deeplearning2/batcher.py

#                                  Apache License
#                            Version 2.0, January 2004
#                         http://www.apache.org/licenses/
#
#    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
#
#    1. Definitions.
#
#       "License" shall mean the terms and conditions for use, reproduction,
#       and distribution as defined by Sections 1 through 9 of this document.
#
#       "Licensor" shall mean the copyright owner or entity authorized by
#       the copyright owner that is granting the License.
#
#       "Legal Entity" shall mean the union of the acting entity and all
#       other entities that control, are controlled by, or are under common
#       control with that entity. For the purposes of this definition,
#       "control" means (i) the power, direct or indirect, to cause the
#       direction or management of such entity, whether by contract or
#       otherwise, or (ii) ownership of fifty percent (50%) or more of the
#       outstanding shares, or (iii) beneficial ownership of such entity.
#
#       "You" (or "Your") shall mean an individual or Legal Entity
#       exercising permissions granted by this License.
#
#       "Source" form shall mean the preferred form for making modifications,
#       including but not limited to software source code, documentation
#       source, and configuration files.
#
#       "Object" form shall mean any form resulting from mechanical
#       transformation or translation of a Source form, including but
#       not limited to compiled object code, generated documentation,
#       and conversions to other media types.
#
#       "Work" shall mean the work of authorship, whether in Source or
#       Object form, made available under the License, as indicated by a
#       copyright notice that is included in or attached to the work
#       (an example is provided in the Appendix below).
#
#       "Derivative Works" shall mean any work, whether in Source or Object
#       form, that is based on (or derived from) the Work and for which the
#       editorial revisions, annotations, elaborations, or other modifications
#       represent, as a whole, an original work of authorship. For the purposes
#       of this License, Derivative Works shall not include works that remain
#       separable from, or merely link (or bind by name) to the interfaces of,
#       the Work and Derivative Works thereof.
#
#       "Contribution" shall mean any work of authorship, including
#       the original version of the Work and any modifications or additions
#       to that Work or Derivative Works thereof, that is intentionally
#       submitted to Licensor for inclusion in the Work by the copyright owner
#       or by an individual or Legal Entity authorized to submit on behalf of
#       the copyright owner. For the purposes of this definition, "submitted"
#       means any form of electronic, verbal, or written communication sent
#       to the Licensor or its representatives, including but not limited to
#       communication on electronic mailing lists, source code control systems,
#       and issue tracking systems that are managed by, or on behalf of, the
#       Licensor for the purpose of discussing and improving the Work, but
#       excluding communication that is conspicuously marked or otherwise
#       designated in writing by the copyright owner as "Not a Contribution."
#
#       "Contributor" shall mean Licensor and any individual or Legal Entity
#       on behalf of whom a Contribution has been received by Licensor and
#       subsequently incorporated within the Work.
#
#    2. Grant of Copyright License. Subject to the terms and conditions of
#       this License, each Contributor hereby grants to You a perpetual,
#       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
#       copyright license to reproduce, prepare Derivative Works of,
#       publicly display, publicly perform, sublicense, and distribute the
#       Work and such Derivative Works in Source or Object form.
#
#    3. Grant of Patent License. Subject to the terms and conditions of
#       this License, each Contributor hereby grants to You a perpetual,
#       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
#       (except as stated in this section) patent license to make, have made,
#       use, offer to sell, sell, import, and otherwise transfer the Work,
#       where such license applies only to those patent claims licensable
#       by such Contributor that are necessarily infringed by their
#       Contribution(s) alone or by combination of their Contribution(s)
#       with the Work to which such Contribution(s) was submitted. If You
#       institute patent litigation against any entity (including a
#       cross-claim or counterclaim in a lawsuit) alleging that the Work
#       or a Contribution incorporated within the Work constitutes direct
#       or contributory patent infringement, then any patent licenses
#       granted to You under this License for that Work shall terminate
#       as of the date such litigation is filed.
#
#    4. Redistribution. You may reproduce and distribute copies of the
#       Work or Derivative Works thereof in any medium, with or without
#       modifications, and in Source or Object form, provided that You
#       meet the following conditions:
#
#       (a) You must give any other recipients of the Work or
#           Derivative Works a copy of this License; and
#
#       (b) You must cause any modified files to carry prominent notices
#           stating that You changed the files; and
#
#       (c) You must retain, in the Source form of any Derivative Works
#           that You distribute, all copyright, patent, trademark, and
#           attribution notices from the Source form of the Work,
#           excluding those notices that do not pertain to any part of
#           the Derivative Works; and
#
#       (d) If the Work includes a "NOTICE" text file as part of its
#           distribution, then any Derivative Works that You distribute must
#           include a readable copy of the attribution notices contained
#           within such NOTICE file, excluding those notices that do not
#           pertain to any part of the Derivative Works, in at least one
#           of the following places: within a NOTICE text file distributed
#           as part of the Derivative Works; within the Source form or
#           documentation, if provided along with the Derivative Works; or,
#           within a display generated by the Derivative Works, if and
#           wherever such third-party notices normally appear. The contents
#           of the NOTICE file are for informational purposes only and
#           do not modify the License. You may add Your own attribution
#           notices within Derivative Works that You distribute, alongside
#           or as an addendum to the NOTICE text from the Work, provided
#           that such additional attribution notices cannot be construed
#           as modifying the License.
#
#       You may add Your own copyright statement to Your modifications and
#       may provide additional or different license terms and conditions
#       for use, reproduction, or distribution of Your modifications, or
#       for any such Derivative Works as a whole, provided Your use,
#       reproduction, and distribution of the Work otherwise complies with
#       the conditions stated in this License.
#
#    5. Submission of Contributions. Unless You explicitly state otherwise,
#       any Contribution intentionally submitted for inclusion in the Work
#       by You to the Licensor shall be under the terms and conditions of
#       this License, without any additional terms or conditions.
#       Notwithstanding the above, nothing herein shall supersede or modify
#       the terms of any separate license agreement you may have executed
#       with Licensor regarding such Contributions.
#
#    6. Trademarks. This License does not grant permission to use the trade
#       names, trademarks, service marks, or product names of the Licensor,
#       except as required for reasonable and customary use in describing the
#       origin of the Work and reproducing the content of the NOTICE file.
#
#    7. Disclaimer of Warranty. Unless required by applicable law or
#       agreed to in writing, Licensor provides the Work (and each
#       Contributor provides its Contributions) on an "AS IS" BASIS,
#       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
#       implied, including, without limitation, any warranties or conditions
#       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
#       PARTICULAR PURPOSE. You are solely responsible for determining the
#       appropriateness of using or redistributing the Work and assume any
#       risks associated with Your exercise of permissions under this License.
#
#    8. Limitation of Liability. In no event and under no legal theory,
#       whether in tort (including negligence), contract, or otherwise,
#       unless required by applicable law (such as deliberate and grossly
#       negligent acts) or agreed to in writing, shall any Contributor be
#       liable to You for damages, including any direct, indirect, special,
#       incidental, or consequential damages of any character arising as a
#       result of this License or out of the use or inability to use the
#       Work (including but not limited to damages for loss of goodwill,
#       work stoppage, computer failure or malfunction, or any and all
#       other commercial damages or losses), even if such Contributor
#       has been advised of the possibility of such damages.
#
#    9. Accepting Warranty or Additional Liability. While redistributing
#       the Work or Derivative Works thereof, You may choose to offer,
#       and charge a fee for, acceptance of support, warranty, indemnity,
#       or other liability obligations and/or rights consistent with this
#       License. However, in accepting such obligations, You may act only
#       on Your own behalf and on Your sole responsibility, not on behalf
#       of any other Contributor, and only if You agree to indemnify,
#       defend, and hold each Contributor harmless for any liability
#       incurred by, or claims asserted against, such Contributor by reason
#       of your accepting any such warranty or additional liability.
#
#    END OF TERMS AND CONDITIONS
#
#    APPENDIX: How to apply the Apache License to your work.
#
#       To apply the Apache License to your work, attach the following
#       boilerplate notice, with the fields enclosed by brackets "{}"
#       replaced with your own identifying information. (Don't include
#       the brackets!)  The text should be enclosed in the appropriate
#       comment syntax for the file format. We also recommend that a
#       file or class name and description of purpose be included on the
#       same "printed page" as the copyright notice for easier
#       identification within third-party archives.
#
#    Copyright {yyyy} {name of copyright owner}
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import numpy as np


class Iterator(object):

    def __init__(self):
        self.batch_index = 0
        self.total_batches_seen = 0

    def reset(self):
        self.batch_index = 0

    def flow_index(self, n, batch_size=32, shuffle=False, seed=None):
        # ensure self.batch_index is 0
        self.reset()
        while 1:
            if seed is not None:
                np.random.seed(seed + self.total_batches_seen)
            if self.batch_index == 0:
                index_array = np.arange(n)
                if shuffle:
                    index_array = np.random.permutation(n)

            current_index = (self.batch_index * batch_size) % n
            if n >= current_index + batch_size:
                current_batch_size = batch_size
                self.batch_index += 1
            else:
                current_batch_size = n - current_index
                self.batch_index = 0
            self.total_batches_seen += 1
            yield (index_array[current_index: current_index + current_batch_size],
                   current_index, current_batch_size)

    def __iter__(self):
        # needed if we want to do something like:
        # for x, y in data_gen.flow(...):
        return self

    def __next__(self, *args, **kwargs):
        return self.next(*args, **kwargs)



import json
import random
import itertools
import os
from collections import namedtuple
from PIL import Image
import bcolz

from brine.schema import Schema
from brine.dataset_manager import DatasetManager
from brine.exceptions import BrineError


def load_dataset(dataset_name, base_path=None):
    """Load a brine dataset that has been installed with `brine install`.

    Parameters
    ----------
    dataset_name : str
        The full name of the dataset to load (eg. examples/cifar10)
    base_path : str
        The path where the dataset was installed. If None, defaults to the current working directory.
    Returns
    -------
    :class:`~brine.dataset.Dataset`
    """
    dataset_manager = DatasetManager.get_from_dir(dataset_name, base_path or os.getcwd())
    return Dataset(dataset_manager)


class Dataset(object):
    """A class to load and interface with a brine dataset

    Use this to get rows from the dataset, create training and validation folds and load images.
    This class also provides methods to create PyTorch and Keras compatible datasets and generators.

    For most cases, use the :func:`~brine.load_dataset` convenience method to load a dataset and return
    a Dataset.
    """

    def __init__(self, dataset_manager, indices=None):
        """
        Parameters
        ----------
        dataset : Dataset
            A brine.Dataset representing the dataset to load.
        indices : list of int
            The indices of the original dataset to use for this dataset. Used for creating folds.
        """
        self.dataset_manager = dataset_manager
        self.indices = indices

        try:
            self.metadata = bcolz.open(os.path.join(self.dataset_manager.path, 'bcolz'), mode='r')
        except IOError:
            raise DatasetError('Dataset %s could not be loaded.' % self.dataset_manager.name)

        self.schema = Schema.from_obj(json.loads(self.metadata.attrs['schema']))
        self.converters = {}
        for i, column in enumerate(self.schema.columns):
            if column.categories:
                if column.isarray():
                    converter = ArrayConverter(CategoryConverter(column.categories))
                    self.converters[i] = converter
                else:
                    converter = CategoryConverter(column.categories)
                    self.converters[i] = converter

        self.extra_data = json.loads(self.metadata.attrs['extra_data'])
        self.Row = namedtuple('Row', [column.name for column in self.schema.columns])
        self.Column = namedtuple('Column', [column.name for column in self.schema.columns])

    def __repr__(self):
        return 'Dataset(name=%s, path=%s)' % (self.dataset_manager.name, self.dataset_manager.path)

    def create_folds(self, fold_sizes, shuffle=False):
        """Returns sub-datasets from this dataset. Useful for creating training and validation folds.

        Parameters
        ----------
        fold_sizes : list of int
            Sizes of the folds to create.
        shuffle : bool
            Whether to shuffle the dataset before creating folds. Defaults to False.

        Returns
        -------
        list of Dataset
            A list of size len(fold_sizes) + 1 of Dataset for the sub-datasets created.
            Any rows leftover will always be returned in the last element of the list as a Dataset.
        """
        total = sum(fold_sizes)
        if total > len(self):
            raise DatasetError
        if self.indices is not None:
            indices = self.indices
        else:
            indices = list(range(0, len(self)))
        if shuffle:
            indices = random.sample(indices, len(self))
        iterator = iter(indices)
        slices = [s for s in (list(itertools.islice(iterator, 0, i)) for i in fold_sizes)]
        remaining = list(iterator)
        slices.append(remaining)
        folds = [Dataset(self.dataset_manager, indices=s) for s in slices]
        return folds

    def to_keras(self, x_column, y_column=None, batch_size=32, processing_function=None, shuffle=True, seed=None):
        """Creates a generator that can be used with Keras' fit_generator.

        Parameters
        ----------
        x_column : str
            The name of the column to treat as the 'x' value
        y_column : str
            The name of the column to treat as the 'y' value. Defaults to None.
        batch_size : int
            The batch size. Defaults to 32.
        processing_function
            Function to apply to each row before it's returned from the generator.
            The function is passed a tuple (x, y) and is expected to return
            (processed_x, processed_y). Images will automatically be converted to numpy arrays
            using Keras' img_to_array.
            y may be None if y_column is None.
            Defaults to None.
        shuffle : bool
            Whether to returned the rows in random order. Defaults to True.
        seed : int
            Seed to use for the random number generator.

        Returns
        -------
        :class:`brine.keras_generator.KerasGenerator`
            A generator that yields batches of samples (x, y) (or just x if y_column is None) from the dataset
            with processing_function applied.
            The generator will continue to yield samples indefinitely.
        """

        try:
            from brine.keras_generator import KerasGenerator
        except ImportError as e:
            if e.name != 'keras':
                raise
            else:
                print('Could not find keras. Please install keras before using this method')
                return None
        return KerasGenerator(self, x_column, y_column,
                              batch_size=batch_size, shuffle=shuffle, processing_function=processing_function,
                              seed=seed)

    def to_pytorch(self, transform=None, transform_columns=None):
        """Creates a PyTorch Dataset that can be passed to a PyTorch DataLoader.

        Parameters
        ----------
        transform : callable
            A callable to apply to each of the rows. If set to None, the default transform will be applied
            which converts any Image columns into PyTorch Tensors.
            Defaults to None.
        transform_columns : 'images' or None
            If transform_columns is set to 'images', each of the Image columns will be passed to the transform
            callable. This is useful to directly apply torchvision transforms on those columns.
            If transform_columns is set to None, the transform function will receive the namedtuple
            representing the entire Row from the Dataset.

        Returns
        -------
        PyTorch Dataset
        """
        try:
            from brine.pytorch_dataset import PytorchDataset
        except ImportError as e:
            if e.name != 'pytorch':
                raise
            else:
                print('Could not find pytorch. Please install pytorch before using this method')
                return None
        return PytorchDataset(self, transform=transform, transform_columns=transform_columns)

    def load_image(self, image_path, imread_fn=None):
        """Get the PIL image for an Image path in the dataset.

        Parameters
        ----------
        image_path : str
            The path to the image in the dataset. This should be retrieved directly from the Row
        imread_fn : callable
            The function to use to read the image. If None, the image will be read with `Image.open`.
            Defaults to None.

        Returns
        -------
        A PIL Image
        """

        p = os.path.join(self.dataset_manager.path, 'images', image_path)
        if imread_fn is None:
            return Image.open(p)
        else:
            return imread_fn(p)

    @property
    def columns(self):
        """Returns a list of Column objects containing information about each column in the dataset
        """
        return self.Column(*self.schema.columns)

    def __enter__(self):
        return self

    def __exit__(self, exc, value, tb):
        pass

    def __getitem__(self, index):
        """
        Returns
        -------
        Row namedtuple
            A namedtuple for the row at the given index.
        """
        if self.indices is None:
            real_index = index
        else:
            real_index = self.indices[index]
        row = list(self.metadata[real_index].item())
        for i, converter in self.converters.items():
            row[i] = converter(row[i])
        return self.Row(*row)

    def __len__(self):
        """
        Returns
        -------
        int
            The length of the entire dataset
        """
        if self.indices is None:
            return self.metadata.shape[0]
        else:
            return len(self.indices)

    def __iter__(self):
        return LoaderIter(self)


class LoaderIter(object):

    def __init__(self, loader):
        self.i = 0
        self.loader = loader

    def __iter__(self):
        return self

    def __next__(self):
        if self.i < len(self.loader):
            self.i = self.i + 1
            return self.loader[self.i - 1]
        else:
            raise StopIteration

    next = __next__


class DatasetError(BrineError):
    pass


class CategoryConverter(object):

    def __init__(self, categories):
        self.categories = categories

    def __call__(self, value):
        return self.categories[value]


class ArrayConverter(object):

    def __init__(self, items_converter):
        self.items_converter = items_converter

    def __call__(self, value):
        return [self.items_converter(x) for x in value]



import codecs
import errno
import glob
import json
import os
import shutil
import re

from brine.env import Env
from brine.exceptions import BrineError


# (?!.*--): cannot contain '--'
# (?!-): cannot start with '-'
# (?<!-): cannot end with '-'
NAME_REGEX = re.compile(r'^(?!.*--)(?!-)([A-Za-z0-9-]+)(?<!-)/(?!-)([A-Za-z0-9-]+)(?<!-)$')


class DatasetManager(object):

    def __init__(self, name, path):
        self.name = name
        self.parse_name(name)
        self.path = os.path.abspath(path)
        self.hidden_file_path = os.path.join(self.path, Env.HIDDEN_FILE_NAME)

    @classmethod
    def list_in_dir(cls, path):
        files = glob.glob(os.path.join(path, Env.DATASETS_DIR_NAME, '*/*', Env.HIDDEN_FILE_NAME))
        files = filter(os.path.isfile, files)
        paths = list(map(os.path.dirname, files))
        names = list(map(lambda p: os.path.relpath(p, start=os.path.join(path, Env.DATASETS_DIR_NAME)), paths))
        dataset_managers = []
        for name, path in zip(names, paths):
            try:
                dataset_manager = cls(name, path)
            except DatasetManagerError:
                continue
            dataset_managers.append(dataset_manager)
        return dataset_managers

    @classmethod
    def get_from_dir(cls, name, path):
        (scope, name_without_scope) = cls.parse_name(name)
        return cls(name, os.path.join(path, Env.DATASETS_DIR_NAME, scope, name_without_scope))

    @staticmethod
    def parse_name(name):
        m = NAME_REGEX.match(name)
        if m is None:
            raise DatasetManagerError('Dataset name %s is not valid.' % name)
        return m.groups()

    def exists(self):
        return os.path.exists(self.hidden_file_path)

    def version(self):
        hidden_file_obj = self._load_hidden_file()
        return hidden_file_obj.get('version')

    def set_version(self, version):
        hidden_file_obj = self._load_hidden_file()
        current_version = hidden_file_obj.get('version')
        if current_version is not None:
            raise DatasetManagerError('Cannot set version on dataset %s.' % self.name)
        hidden_file_obj['version'] = version
        self._save_hidden_file(hidden_file_obj)

    def _load_hidden_file(self):
        try:
            with codecs.open(self.hidden_file_path, 'r', encoding='utf-8') as f:
                return json.loads(f.read())
        except (IOError, ValueError):
            return {}

    def _save_hidden_file(self, hidden_file_obj):
        try:
            with codecs.open(self.hidden_file_path, 'w', encoding='utf-8') as f:
                f.write(json.dumps(hidden_file_obj))
        except (IOError, ValueError):
            raise DatasetManagerError('Could not save file %s' % self.hidden_file_path)

    def remove(self):
        if not self.exists():
            raise DatasetManagerError('Dataset %s is not installed.' % self.name)
        try:
            shutil.rmtree(self.path)
        except OSError as ex:
            if ex.errno != errno.ENOENT:
                raise DatasetManagerError('Dataset %s could not be removed.' % self.name)

    def check_can_install(self):
        if self.exists():
            raise DatasetManagerError('Dataset %s is already installed.' % self.name)

        if os.path.exists(self.path):
            raise DatasetManagerError('Dataset %s directory could not be created at %s.' % (self.name, self.path))

        p = self.path
        while not os.path.exists(p):
            p = os.path.dirname(p)

        if not os.path.isdir(p) or not os.access(p, os.R_OK | os.W_OK | os.F_OK):
            raise DatasetManagerError('Dataset %s directory could not be created at %s.' % (self.name, self.path))

    def check_can_push(self):
        if not self.exists():
            raise DatasetManagerError('Dataset %s has not been built.' % self.name)

        if self.version() is not None:
            raise DatasetManagerError('Dataset %s has already been pushed.' % self.name)

    def create_from_dir(self, source_path, version=None):
        if self.exists():
            raise DatasetManagerError('Dataset %s is already installed.' % self.name)

        dataset_manager = DatasetManager(self.name, source_path)
        dataset_manager.set_version(version)

        path_parent = os.path.dirname(self.path)
        try:
            os.makedirs(path_parent)
        except OSError as ex:
            if ex.errno != errno.EEXIST or not os.path.isdir(path_parent):
                raise DatasetManagerError('Dataset %s directory could not be created at %s.' % (self.name, self.path))

        try:
            shutil.move(source_path, self.path)
        except OSError as ex:
            raise DatasetManagerError('Dataset %s directory could not be created from %s.' % (self.name, source_path))


class DatasetManagerError(BrineError):
    pass



import json
from collections import OrderedDict

from brine.api import get_dataset_for_info


def info(dataset_name):
    response = get_dataset_for_info(dataset_name)
    ordered = OrderedDict()
    ordered['name'] = response['name']
    ordered['description'] = response['description']
    ordered['versions'] = response['versions']

    print(json.dumps(ordered, indent=4))



from setuptools import setup, find_packages


setup(
    name='brine-io',
    description='Dataset management for computer vision',
    long_description='Dataset management for computer vision',
    version='0.5.0',
    url='https://github.com/brine-io/brine',
    download_url='https://github.com/brine-io/brine/archive/0.5.0.tar.gz',
    author='David Ye',
    author_email='david@brine.io',
    keywords='brine',
    license='LICENSE',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    packages=find_packages(),
    install_requires=[
        "requests>=2.14.0",
        "tqdm>=4.17.0",
        "numpy>=1.7.0",
        "bcolz>=1.1.1",
        "pandas>=0.19.0",
        "pillow",
    ],
    entry_points={
        'console_scripts': [
            'brine=brine.__main__:main',
        ],
    },
    python_requires='>=3',
)



import os
from brine.dataset_manager import DatasetManager


def uninstall(dataset_name):
    dataset_manager = DatasetManager.get_from_dir(dataset_name, os.getcwd())
    version = dataset_manager.version()
    dataset_manager.remove()

    print('Dataset %s (v%s) was uninstalled.' % (dataset_name, version))
